
#ifndef ThereminoFilters_h
#define ThereminoFilters_h

#include <Arduino.h>

class Filter 
{
  private:
  float tauUS;       // filter time constant in US
  uint32_t lastUS;   // last time in microseconds
  float outValue;

  public:  
  bool  isHighPass;

  Filter(float freq, bool hiPass) 
  {
    outValue = 0;
    lastUS = micros();
    isHighPass = hiPass;
    setFrequency(freq);
  }

  void setFrequency(float freq)
  {
    tauUS = 1e6 / (TWO_PI * freq);
  }
  
  float run(float Value) 
  { 
    float filterData = Value;
    
    // the time rollover is corrected automatically by the unsigned math
    uint32_t time = micros();
    uint32_t elapsedUS = time - lastUS;
    lastUS = time;
          
    float tauSamps = tauUS / elapsedUS;
    float ampFactor = exp(-1.0 / tauSamps);

    outValue = (1.0 - ampFactor) * filterData + ampFactor * outValue;
    
    return isHighPass ? filterData - outValue : outValue;
  }
};

#endif
