//================================================================================================
//                                  cIotModuleCom header
//================================================================================================

class cIotModuleCom
{
  private:

    static const uint16_t UDP_COM_BUF_LEN = 1416;
    static const uint16_t FUOTA_TCP_DATA_BUF_LEN = 1460;
    static const uint16_t ARDUINO_IDE_UDP_PORT_NUMBER = 3232;
    static const uint16_t COM_TASK_STACK_SIZE = 2 * 4096;

    enum FUNC_ID : uint8_t
    {
      SCAN,
      PROG,
      GETSINGLE,
      XCHG,
      NAME,
      EXEC,
      PWM,
      STEP,
      ADC24,
    };
    //  do not remove pragma! it's mandatory because of RPC stuff!
#pragma pack(push, 1)
    typedef struct sHeader
    {
      char Protocol[6];
      uint16_t ProtocolVer;
      char Who;
      uint16_t FirmwareVer;
      FUNC_ID FuncId;
    } tHeader;

    typedef struct sIotHALData
    {
      tHeader Header;
      union
      {
        struct
        {
          uint8_t MAC[6];
          uint8_t ModuleNameSize;
          uint8_t ModuleName[32];
          uint8_t PinoutType;
          uint8_t ProgStatus;
          uint8_t TotalPinsNumber;
          uint8_t PinName[IOTMODULE_TOTAL_PINS];
        } ScanReply;
        struct
        {
          uint8_t TotalPinsNumber;
          uint8_t PinType[IOTMODULE_TOTAL_PINS];
        } ProgReq;
        struct
        {
          uint8_t ProgStatus;
        } ProgReply;
        struct
        {
          uint8_t PinIndex;
        } GetSingleReq;
        struct
        {
          uint8_t PinName;
          uint8_t TotalPinTypes;
          uint8_t PinType[cIotModuleApp::PIN_TYPE_SIZE];
        } GetSingleReply;
        struct
        {
          uint8_t TotalOutValues;
          float OutValue[IOTMODULE_TOTAL_PINS];
        } XchgReq;
        struct
        {
          uint8_t ProgStatus;
          uint8_t TotalInValues;
          float InValue[IOTMODULE_TOTAL_PINS];
        } XchgReply;
        struct
        {
          uint8_t ModuleNameSize;
          uint8_t ModuleName[32];
        } NameReq;
        struct
        {
          uint8_t ModuleNameSize;
          uint8_t ModuleName[32];
        } NameReply;
        struct
        {
          uint8_t PinIndex;
          float Freq;
        } PwmReq;
        struct
        {
          bool ReceivedOK;
        } PwmReply;
        struct
        {
          uint8_t PinIndex;
          uint16_t StepperSpeedHz;
          uint16_t StepperAccHz_per_mS;
        } StepReq;
        struct
        {
          bool ReceivedOK;
        } StepReply;
        struct
        {
          uint16_t Sps;
          bool MultipleSetups;
          bool FiltersSingleCycle;
          uint8_t PreFilterType;
          uint8_t PostFilterType;
          uint16_t ActiveChannels;
          uint16_t BiasedChannels;
          uint16_t DiffChannels;
          uint16_t PseudoChannels;
          uint8_t ChannelsGains[3];
        } Adc24Req;
        struct
        {
          bool ReceivedOK;
        } Adc24Reply;
      } Body;
    } tIotHALData;
#pragma pack(pop)

    wifi_config_t WifiConfig;
    TaskHandle_t hComServerTask;
    TaskHandle_t hFUOTAWatchDogTask;
    TimerHandle_t hTimerCallback;
    EventGroupHandle_t hEventGroup;
    pWifiEventHandler WifiEventHandlerCallback;
    int hIotHALSocket;
    uint8_t WiFiMAC[6];
    uint16_t FramesPerSecond;
    uint32_t TaskInactiveMicroSec;
    uint32_t TaskActiveMicroSec;
    uint8_t IO_WatchDogCounter;
    bool FUOTA_WatchDogEnable;
    uint16_t FUOTA_WatchDogCounter;

    static esp_err_t WifiEventHandler(void *ctx, system_event_t *event);
    static void ComServerTask(void *pvParameters);
    static void TimerCallback(TimerHandle_t xExpiredTimer);
    void Fuota(void);
    void NetModuleRPC(const void *pComBuf, int *pIotHALDataLen);

  public:

    //  IotModule Application class
    cIotModuleApp IotModuleApp;

    cIotModuleCom();
    void Initialize(pWifiEventHandler WifiEventHandler);
    void GetTaskStats(uint16_t *pFramesPerSec, uint16_t *pTaskActiveMicroSec, uint16_t *pTaskInactiveMicroSec);
};

