//================================================================================================
//                                  cIotModuleApp body
//================================================================================================

cIotModuleApp::cIotModuleApp()
{
}

void cIotModuleApp::Initialize(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  char pName[32 + 1];
  GetModuleName(pName, sizeof(pName));
  //  adc calibration
  if (esp_adc_cal_check_efuse(ESP_ADC_CAL_VAL_EFUSE_TP) == ESP_OK)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration - eFuse two point supported", __FILENAME__, __LINE__);
  }
  else if (esp_adc_cal_check_efuse(ESP_ADC_CAL_VAL_EFUSE_VREF) == ESP_OK)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration - eFuse Vref supported", __FILENAME__, __LINE__);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration not supported", __FILENAME__, __LINE__);
  }
  //  set attenuation
  analogSetAttenuation(ADC_11db);
  //  Characterize ADC at particular attenuation
  esp_adc_cal_value_t val_type = esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 0, &AdcChars);
  //  Check type of calibration value used to characterize ADC
  if (val_type == ESP_ADC_CAL_VAL_EFUSE_TP)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration done with eFuse two point", __FILENAME__, __LINE__);
  }
  else if (val_type == ESP_ADC_CAL_VAL_EFUSE_VREF)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration done with eFuse Vref", __FILENAME__, __LINE__);
  }

  //  initialize ISR service for GPIO
  ESP_ERROR_CHECK(gpio_install_isr_service(ESP_INTR_FLAG_LEVEL1));
  //  uninstall is never used
  //  gpio_uninstall_isr_service();

  //  IotModuleOth initialization
  IotModuleOth.Initialize();

  //  IotModuleAdc initialization
  IotModuleAdc.Initialize();

  //  start count total active pins
  AllPinsData.TotalPinsNumber = 0;

  //  real pins number
  uint8_t InOutPins = IOTMODULE_REAL_PINS;
  //  if less pins wanted
  if (IOTMODULE_INOUT_PINS < InOutPins)
  {
    //  less pins used
    InOutPins = IOTMODULE_INOUT_PINS;
  }

  //  if at least one adc channel is defined
  if (IOTMODULE_ADC24_PINS > 0)
  {
    //  set AD7124 control pins
    for (uint8_t i = 0; (i < IOTMODULE_REAL_PINS); i++)
    {
      //  if this pin is reserved for adc24
      if ( ModuleProps.PinsProps[i].Name == AD7124_PIN_NUM_MOSI ||
           ModuleProps.PinsProps[i].Name == AD7124_PIN_NUM_MISO ||
           ModuleProps.PinsProps[i].Name == AD7124_PIN_NUM_CLK)
      {
        //  set ADC24 GPIO parameters
        switch(ModuleProps.PinsProps[i].Name)
        {
          case AD7124_PIN_NUM_MOSI:
            IotModuleAdc.SetSPI_Control_Pin(cIotModuleAdc::AD7124_MOSI, (gpio_num_t)ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].ResetMode);
            break;
          case AD7124_PIN_NUM_MISO:
            IotModuleAdc.SetSPI_Control_Pin(cIotModuleAdc::AD7124_MISO, (gpio_num_t)ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].ResetMode);
            break;
          case AD7124_PIN_NUM_CLK:
            IotModuleAdc.SetSPI_Control_Pin(cIotModuleAdc::AD7124_CLK, (gpio_num_t)ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].ResetMode);
            break;
        }
        //  set ADC24 GPIO parameters
        switch(ModuleProps.PinsProps[i].Name)
        {
          case AD7124_PIN_NUM_MOSI:
          case AD7124_PIN_NUM_MISO:
          case AD7124_PIN_NUM_CLK:
            //  leave only generic type
            ModuleProps.PinsProps[i].Channel1 = GenericPinProps.Channel1;
            ModuleProps.PinsProps[i].CapName = GenericPinProps.CapName;
            ModuleProps.PinsProps[i].ResetMode = GenericPinProps.ResetMode; //todo
            ModuleProps.PinsProps[i].TypesSize = GenericPinProps.TypesSize;
            //  copy the generic pin types
            for (uint8_t j = 0; j < GenericPinProps.TypesSize; j++)
            {
              ModuleProps.PinsProps[i].Types[j] = GenericPinProps.Types[j];
            }
            break;
        }
      }
    }
  }
  AllPinsData.TotalPinsNumber += InOutPins;

  //  if at least one adc channel is defined
  if (IOTMODULE_ADC24_PINS > 0)
  {
    uint8_t ADC24VirtPinName = FIRST_AD7124_PIN_NAME;
    uint8_t ADC24VirtPinCount = IOTMODULE_ADC24_PIN_MASTER + IOTMODULE_ADC24_PINS;
    //  adc24 master pin
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Name = ADC24VirtPinName++;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Channel1 = ADC24_MasterPinProps.Channel1;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].CapName = ADC24_MasterPinProps.CapName;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].ResetMode = ADC24_MasterPinProps.ResetMode;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].TypesSize = ADC24_MasterPinProps.TypesSize;
    //  copy the pin types
    for (uint8_t i = 0; i < ADC24_MasterPinProps.TypesSize; i++)
    {
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Types[i] = ADC24_MasterPinProps.Types[i];
    }
    ADC24VirtPinCount--;
    AllPinsData.TotalPinsNumber++;
    //  adc24 data pins
    for (; ADC24VirtPinCount != 0; ADC24VirtPinCount--)
    {
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Name = ADC24VirtPinName++;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Channel1 = ADC24ChannelPinProps.Channel1;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].CapName = ADC24ChannelPinProps.CapName;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].ResetMode = ADC24ChannelPinProps.ResetMode;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].TypesSize = ADC24ChannelPinProps.TypesSize;
      //  copy the pin types
      for (uint8_t i = 0; i < ADC24ChannelPinProps.TypesSize; i++)
      {
        ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Types[i] = ADC24ChannelPinProps.Types[i];
      }
      AllPinsData.TotalPinsNumber++;
    }
  }

  //  if at least one generic pin is defined
  if (IOTMODULE_GENERIC_PINS > 0)
  {
    //  compile the props for the generic pins defined
    uint8_t GenVirtPinName = FIRST_GENERIC_PIN_NAME;
    uint8_t GenVirtPinCount = IOTMODULE_GENERIC_PINS;
    for (; GenVirtPinCount != 0; GenVirtPinCount--)
    {
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Name = GenVirtPinName++;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Channel1 = GenericPinProps.Channel1;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].CapName = GenericPinProps.CapName;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].ResetMode = GenericPinProps.ResetMode;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].TypesSize = GenericPinProps.TypesSize;
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TypesSize : %d", __FILENAME__, __LINE__, GenericPinProps.TypesSize);
      //  copy the pin types
      for (uint8_t i = 0; i < GenericPinProps.TypesSize; i++)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Types : %d", __FILENAME__, __LINE__, GenericPinProps.Types[i]);
        ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Types[i] = GenericPinProps.Types[i];
      }
      AllPinsData.TotalPinsNumber++;
    }
  }

  //  reset floating values and pin type
  for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
  {
    AllPinsData.PinsParams[Index].fValue = 0.0f;
    AllPinsData.PinsParams[Index].Type = UNUSED;
  }
  //  initialize pin->index lookup table
  for (uint8_t Index = 0; Index < IOTMODULE_LOOKUP_TABLE_SIZE; Index++)
  {
    //  set all index values to 255 == not valid
    PinIndexLookupTable[Index] = 255;
  }
  //  build pin->index lookup table
  for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
  {
    PinIndexLookupTable[ModuleProps.PinsProps[Index].Name] = Index;
  }

  //  module is not programmed
  ProgStatus = NOTHING;

  //  display pinout type
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_PINOUT_TYPE : %d", __FILENAME__, __LINE__, IOTMODULE_PINOUT_TYPE);
  //  display I/O module pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_INOUT_PINS : %d", __FILENAME__, __LINE__, IOTMODULE_INOUT_PINS);
  //  display adc24 pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_ADC24_PINS : %d", __FILENAME__, __LINE__,  IOTMODULE_ADC24_PIN_MASTER + IOTMODULE_ADC24_PINS);
  //  display generic pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_GENERIC_PINS : %d", __FILENAME__, __LINE__, IOTMODULE_GENERIC_PINS);
  //  display total handled pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal TotalPinsNumber : %d", __FILENAME__, __LINE__, AllPinsData.TotalPinsNumber);
  //  display pin list
  for (uint8_t i = 0; i < AllPinsData.TotalPinsNumber; i++)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Pin List , Index : %d , Name : %d, TypesSize : %d", __FILENAME__, __LINE__, i, ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].TypesSize);
    for (uint8_t j = 0; j < ModuleProps.PinsProps[i].TypesSize; j++)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Pin List , Type : %d", __FILENAME__, __LINE__, ModuleProps.PinsProps[i].Types[j]);
    }
  }
}

void cIotModuleApp::ADC24SetParam(const uint16_t Sps, const bool MultipleSetups, const bool FiltersSingleCycle, const uint8_t PreFilterType,
                                  const uint8_t PostFilterType, const uint16_t ActiveChannels, const uint16_t BiasedChannels, const uint16_t DiffChannels,
                                  const uint16_t PseudoChannels, const uint8_t ChannelsGains0, const uint8_t ChannelsGains1, const uint8_t ChannelsGains2)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - ADC24SetParam(App) called", __FILENAME__, __LINE__);
  //  call IotModuleAdc.ADC24SetParam
  IotModuleAdc.SetParam(Sps, MultipleSetups, FiltersSingleCycle, PreFilterType,
                        PostFilterType, ActiveChannels, BiasedChannels, DiffChannels,
                        PseudoChannels, ChannelsGains0, ChannelsGains1, ChannelsGains2);
  //  set AD7124 section as programmed
  ProgStatus = static_cast<PROG_STATUS>(ProgStatus | AD7124);
}

void cIotModuleApp::PreFuota(void)
{
  //  Stop any interrupt activity
  IotModuleAdc.PreFuota();
  IotModuleOth.PreFuota();
}

void cIotModuleApp::SetModuleName(char *pName)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - name : %s", __FILENAME__, __LINE__, pName);

  nvs_handle my_nvshandle;
  esp_err_t err = nvs_open_from_partition("nvs", "IotModule", NVS_READWRITE, &my_nvshandle);
  //esp_err_t err = nvs_open("IotModule", NVS_READWRITE, &my_nvshandle);
  if (err == ESP_OK)
  {
    err = nvs_set_str(my_nvshandle, "ModuleName", pName);
    if (err != ESP_OK)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_set_str error", __FILENAME__, __LINE__);
    }
    nvs_close(my_nvshandle);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_open error", __FILENAME__, __LINE__);
  }
}

void cIotModuleApp::GetModuleName(char *pName, const uint8_t NameSize)
{
  char pDefaultName[] = "NoName";
  nvs_handle my_nvshandle;
  esp_err_t err = nvs_open_from_partition("nvs", "IotModule", NVS_READWRITE, &my_nvshandle);
  //esp_err_t err = nvs_open("IotModule", NVS_READWRITE, &my_nvshandle);
  if (err == ESP_OK)
  {
    size_t Size = NameSize;
    err = nvs_get_str(my_nvshandle, "ModuleName", pName, &Size);
    if (err != ESP_OK)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_get_str error", __FILENAME__, __LINE__);
      strcpy((char*)pName, pDefaultName);
      err = nvs_set_str(my_nvshandle, "ModuleName", pName);
      if (err != ESP_OK)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_set_str error", __FILENAME__, __LINE__);
      }
    }
    nvs_close(my_nvshandle);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_open error", __FILENAME__, __LINE__);
  }
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - name : %s", __FILENAME__, __LINE__, pName);
  if (err != ESP_OK)
  {
    strcpy((char*)pName, pDefaultName);
  }
}

uint8_t IRAM_ATTR cIotModuleApp::GetOutputPinsNumber(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  return AllPinsData.OutputPinsNumber;
}

uint8_t IRAM_ATTR cIotModuleApp::GetInputPinsNumber(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  return AllPinsData.InputPinsNumber;
}

char IRAM_ATTR cIotModuleApp::GetProgStatus(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  char retval = 'N';
  if ((ProgStatus & MODULE) == MODULE)
  {
    retval = 'Y';
  }
  // if((IOTMODULE_ADC24_PIN_MASTER == 0) && (ProgStatus == MODULE))
  // {
  // retval = 'Y';
  // }
  // else if((IOTMODULE_ADC24_PIN_MASTER == 1) && (ProgStatus == (MODULE | AD7124)))
  // {
  // retval = 'Y';
  // }
  return retval;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function update the output values on the esp32 hardware
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::SetOutputValues(bool Reset)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  if (GetProgStatus() == 'Y')
  {
    for (uint8_t IndexCount = 0; IndexCount < AllPinsData.OutputPinsNumber; IndexCount++)
    {
      uint8_t Index = AllPinsData.OutputIndexes[IndexCount];
      if (Reset)
      {
        AllPinsData.PinsParams[Index].fValue = 0.0f;
      }
      switch (AllPinsData.PinsParams[Index].Type)
      {
        case DIG_OUT:
          ESP_ERROR_CHECK(gpio_set_level((gpio_num_t)AllPinsData.PinsParams[Index].Name, AllPinsData.PinsParams[Index].fValue == 0.0f ? 0 : 1));
          break;
        case PWM:
          // here the fValue is a number from 0.0f to 1.0f representing the PWM pulse from min to max
          // the fValue is sent directly to the function PwmSetWidth that accepts a value from 0.0f to 1.0f representing the PWM pulse from min to max
          PwmSetWidth(AllPinsData.PinsParams[Index].PwmChannel, AllPinsData.PinsParams[Index].fValue);
          break;
        case SERVO:
          // here the fValue is microseconds from 0 to 4000
          // the SERVO pins are fixed to 50 Hz so the total time is 20 mS (20000 uS)
          // the fValue is divided by 20000 for the function PwmSetWidth that accepts a value from 0.0f to 1.0f representing the PWM pulse from min to max
          PwmSetWidth(AllPinsData.PinsParams[Index].PwmChannel, AllPinsData.PinsParams[Index].fValue / 20000.0f);
          break;
        case DAC_8:
          ESP_ERROR_CHECK(dac_output_voltage((dac_channel_t)ModuleProps.PinsProps[Index].Channel1, (uint8_t)(AllPinsData.PinsParams[Index].fValue)));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - DAC_8 - Index : %d, channel : %d , fValue : %f", __FILENAME__, __LINE__, Index, ModuleProps.PinsProps[Index].Channel1, AllPinsData.PinsParams[Index].fValue);
          break;
        case GEN_OUT_8:
        case GEN_OUT_16:
        case GEN_OUT_24:
        case GEN_OUT_FLOAT:
          // generic output values are written by genericWrite* functions
          break;
        default:
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not handled : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].Type);
          break;
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function update the input values of the esp32 hardware
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::GetInputValues(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  //  if module programmed
  if (GetProgStatus() == 'Y')
  {
    //  if CAP pin configured cycle between them every time this function is requested
    if (AllPinsData.CapNumber > 1)
    {
      AllPinsData.CapSeqCounter++;
      if (AllPinsData.CapSeqCounter == AllPinsData.CapNumber)
      {
        AllPinsData.CapSeqCounter = 0;
      }
    }
    uint8_t AdcChannelIndex = 0;
    uint16_t AdcActiveChannels = IotModuleAdc.GetActiveChannels();
    //  for all the input pins
    for (uint8_t IndexCount = 0; IndexCount < AllPinsData.InputPinsNumber; IndexCount++)
    {
      uint8_t Index = AllPinsData.InputIndexes[IndexCount];
      switch (AllPinsData.PinsParams[Index].Type)
      {
        case DIG_IN:
        case DIG_IN_PU:
          AllPinsData.PinsParams[Index].fValue = gpio_get_level((gpio_num_t)AllPinsData.PinsParams[Index].Name) == 0 ? 0.0 : 1.0;
          break;
        case ADC_16:
          {
            uint32_t tmp = 0;
            for (uint8_t i = 0; i < ADC_CONVERSION_OVERSAMPLING; i++)
            {
              tmp += esp_adc_cal_raw_to_voltage(analogRead(AllPinsData.PinsParams[Index].Name), (const esp_adc_cal_characteristics_t *)&AdcChars);
            }
            AllPinsData.PinsParams[Index].fValue = (float)tmp * adcMult + adcShift;
          }
          break;
        case CAP_16:
          // ----------------------------------------------------------------------------------------------
          // Correction 1
          // - Sometimes there are "glitches" to zero or low values
          // - Each channel is read three times and the highest value is used (two times is not enought)
          // ----------------------------------------------------------------------------------------------
          // Correction 2
          // - The Pin32 or Pin33 used alone (with no other CAP_16 configured) is always "1"
          // - When reading as CAP_16 a single Pin32 we read the Pin33 before
          // - When reading as CAP_16 a single Pin33 we read the Pin32 before
          // ----------------------------------------------------------------------------------------------
          // The touchRead function is slow (460 uS), only one pin at a time is read every xchg cycle
          // It would be better but uselessly complex to prepare all the values in a separate thread
          // Is better to use a custom function that simply WORKS and is faster than this, if exist
          // ----------------------------------------------------------------------------------------------
          if (AllPinsData.CapSeqCounter == AllPinsData.PinsParams[Index].CapSeq)
          {
            uint8_t n1;
            uint8_t n2;
            uint8_t name;
            name = ModuleProps.PinsProps[Index].CapName;
            if (AllPinsData.CapNumber == 1)
            {
              if (name == 32) n1 = touchRead(33);
              if (name == 33) n1 = touchRead(32);
            }
            n1 = touchRead(name);
            n2 = touchRead(name);
            if (n2 > n1) n1 = n2;
            n2 = touchRead(name);
            if (n2 > n1) n1 = n2;
            AllPinsData.PinsParams[Index].fValue = n1 * 256.0f;
          }
          break;
        case COUNTER:
        case COUNTER_PU:
          {
            int16_t Count;
            ESP_ERROR_CHECK(pcnt_get_counter_value((pcnt_unit_t)AllPinsData.PinsParams[Index].Channel2, &Count));
            AllPinsData.PinsParams[Index].fValue = (uint16_t)Count;
          }
          break;
        case PERIOD:
        case PERIOD_PU:
          {
            uint32_t Count;
            IotModuleOth.PeriodGetCount(AllPinsData.PinsParams[Index].Channel2, &Count);
            AllPinsData.PinsParams[Index].fValue = Count;
          }
          break;
        case ENCODER_A:
        case ENCODER_A_PU:
          {
            int16_t Count;
#if IOTMODULE_ENCODERS_TYPE == 0
            ESP_ERROR_CHECK(pcnt_get_counter_value((pcnt_unit_t)AllPinsData.PinsParams[Index].Channel2, &Count));
#elif IOTMODULE_ENCODERS_TYPE == 1
            IotModuleOth.SWQEncoderGetQuadratureCount(AllPinsData.PinsParams[Index].Channel2, &Count);
#else
#error "IOTMODULE_ENCODERS_TYPE IS NOT DEFINED IN IOTMODULESETUP.H"
#endif
            AllPinsData.PinsParams[Index].fValue = (uint16_t)Count;
          }
          break;
        case ADC_24:
          AllPinsData.PinsParams[Index].fValue = IotModuleAdc.GetISR_Counter();
          break;
        case ADC_24_CH:
          //  if at least there is one channel active
          if (AdcActiveChannels)
          {
            //  while not active channels
            while (!(AdcActiveChannels & 0x1))
            {
              //  skip not active channels  
              AdcActiveChannels >>= 1;
              AdcChannelIndex++;
            }
          }
          //  get integer conversion value for this active channel converting it to float
          AllPinsData.PinsParams[Index].fValue = IotModuleAdc.GetISR_Channel_uint32_Value(AdcChannelIndex++);
          AdcActiveChannels >>= 1;
          break;
        //  generic input values are written by genericRead(*) functions
        case GEN_IN_8:
        case GEN_IN_16:
        case GEN_IN_24:
        case GEN_IN_FLOAT:
          break;
        //  not implemented
        default:
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not implemented : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].Type);
          break;
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                     RPC functions
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCSetModuleName(const uint8_t ModuleNameSize, uint8_t *pModuleName)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  char pName[32 + 1];
  strncpy((char*)pName, (char*)pModuleName, ModuleNameSize);
  pName[ModuleNameSize] = '\0';
  SetModuleName(pName);
}

void cIotModuleApp::RPCGetModuleName(uint8_t *pModuleNameSize, uint8_t *pModuleName)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  char pName[32 + 1];
  GetModuleName(pName, sizeof(pName));
  uint8_t tmp;
  tmp = strlen(pName);
  *pModuleNameSize = tmp;
  strncpy((char*)pModuleName, pName, tmp);
}

uint8_t cIotModuleApp::RPCGetTotalPinsNumber(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - called, TotalPinsNumber : %d", __FILENAME__, __LINE__, AllPinsData.TotalPinsNumber);
  return AllPinsData.TotalPinsNumber;
}

uint8_t cIotModuleApp::RPCGetPinName(const uint8_t PinIndex)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - called, PinIndex : %d, Name : %d", __FILENAME__, __LINE__, PinIndex, ModuleProps.PinsProps[PinIndex].Name);
  return ModuleProps.PinsProps[PinIndex].Name;
}

void cIotModuleApp::RPCGetScanPinTypes(const uint8_t PinIndex, uint8_t *pRPCBuffer)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - called, PinIndex : %d", __FILENAME__, __LINE__, PinIndex);
  memcpy(pRPCBuffer, &ModuleProps.PinsProps[PinIndex].Types, ModuleProps.PinsProps[PinIndex].TypesSize);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function get all the possible pin types for a certain pin and removes some types based on the
//  remaining resources of the moment
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCGetDynamicPinTypes(const uint8_t PinIndex, uint8_t *pTotalPinTypes, uint8_t *pPinType)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called - PinIndex : %d", __FILENAME__, __LINE__, PinIndex);

  if (PinIndex < AllPinsData.TotalPinsNumber)
  {
    uint8_t Index = 0;
    bool Found = false;

    if (PinIndex > 0)
    {
      //  if previous pin is encodera* then this pin must be only encoderb* or unused
      if (AllPinsData.PinsParams[PinIndex - 1].Type == ENCODER_A)
      {
        pPinType[Index++] = UNUSED;
        pPinType[Index++] = ENCODER_B;
        Found = true;
      }
      else if (AllPinsData.PinsParams[PinIndex - 1].Type == ENCODER_A_PU)
      {
        pPinType[Index++] = UNUSED;
        pPinType[Index++] = ENCODER_B_PU;
        Found = true;
      }
    }
    //  otherwise
    if (!Found)
    {
      //  for all the possible pin types of this pin
      for (uint8_t Index1 = 0; Index1 < ModuleProps.PinsProps[PinIndex].TypesSize; Index1++)
      {
        PIN_TYPE Type = ModuleProps.PinsProps[PinIndex].Types[Index1];
        switch (Type)
        {
          //  add servo or pwm to pin types:
          //  - if ledc channels are available or
          //  - if they may be confirmed because already configured
          case SERVO:
          case PWM:
            if ((LedcChannelNumber < MAX_LEDC_CHANNELS) || (AllPinsData.PinsParams[PinIndex].Type == SERVO) || (AllPinsData.PinsParams[PinIndex].Type == PWM))
            {

              // LIVIO - GRRRRRR
              // ---------------------------------------------------------------------------------
              //  If Servo channels count is odd we loose a channel.
              //  So if you set seven Servos, then you can not set nine Pwm but only eight.
              //  And if you set fifteen Servos the sixteenth can not be set as Pwm.
              // ---------------------------------------------------------------------------------
              //  Bisogna stare attenti anche quando si aggiunge un Servo !!!
              //  perché si può finire nella situazione non permessa
              // ---------------------------------------------------------------------------------
              //  uint8_t servoCount = ServoChannelNumber;
              //  if (AllPinsData.PinsParams[PinIndex].Type == SERVO) servoCount -= 1;
              ////  if (Type != PWM || servoCount % 2 == 0 || ServoChannelNumber + PwmChannelNumber < MAX_LEDC_CHANNELS -1)
              //  if (servoCount % 2 == 0 || ServoChannelNumber + PwmChannelNumber < MAX_LEDC_CHANNELS -1)
              //  {
              //    pPinType[Index++] = Type;
              //  }
              // ---------------------------------------------------------------------------------

              pPinType[Index++] = Type;

            }
            break;
          //  add counter* to pin types:
          //  - if counters channels are available or
          //  - if they may be confirmed because already configured
          case COUNTER:
          case COUNTER_PU:
            if ((CounterChannelNumber < MAX_COUNTER_CHANNELS) || (AllPinsData.PinsParams[PinIndex].Type == COUNTER) || (AllPinsData.PinsParams[PinIndex].Type == COUNTER_PU))
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add encodera* to pin types:
          //  - if counters channels are available or
          //  - if they may be confirmed because already configured
          case PERIOD:
          case PERIOD_PU:
            if ((PeriodChannelNumber < cIotModuleOth::MAX_PERIOD_CHANNELS) || (AllPinsData.PinsParams[PinIndex].Type == PERIOD) || (AllPinsData.PinsParams[PinIndex].Type == PERIOD_PU))
            {
              pPinType[Index++] = Type;
            }
            break;
          case ENCODER_A:
          case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
            if ((CounterChannelNumber < MAX_COUNTER_CHANNELS) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_A) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_A_PU))
#elif IOTMODULE_ENCODERS_TYPE == 1
            if ((SWQEncoderChannelNumber < cIotModuleOth::MAX_SWQENCODER_CHANNELS) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_A) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_A_PU))
#endif
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add encoderb* to pin types:
          //  - if the actual pin is encodera* and the next pin is not encoderb*
          //  - if they may be confirmed because already configured
          case ENCODER_B:
          case ENCODER_B_PU:
            if ((((AllPinsData.PinsParams[PinIndex].Type == ENCODER_A) && (AllPinsData.PinsParams[PinIndex + 1].Type != ENCODER_B)) &&
                 ((AllPinsData.PinsParams[PinIndex].Type == ENCODER_A_PU) && (AllPinsData.PinsParams[PinIndex + 1].Type != ENCODER_B_PU))) ||
                (AllPinsData.PinsParams[PinIndex].Type == ENCODER_B) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_B_PU))
            {
              pPinType[Index++] = (uint8_t)Type;
            }
            break;
          //  add ADC_24_CH_B only if previous is already programmed as ADC_24_CH and the index is odd
          case ADC_24_CH_B:
            if ((AllPinsData.PinsParams[PinIndex - 1].Type == ADC_24_CH) && (IOTMODULE_ADC24_PINS % 2 == 0))
            {
              pPinType[Index++] = (uint8_t)Type;
            }
            break;
          //  add always the remaining types
          default:
            pPinType[Index++] = (uint8_t)Type;
            break;
        }
      }
    }
    //  Set number of valid pin types
    *pTotalPinTypes = Index;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - TotalPinTypes : %d", __FILENAME__, __LINE__, Index);
    for (uint8_t i = 0; i < Index; i++)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Type : %d", __FILENAME__, __LINE__, pPinType[i]);
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin number mismatch", __FILENAME__, __LINE__);
    *pTotalPinTypes = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function programs the esp32 hardware to get the pins ready to handle the programmed function
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCSetAllPinsParams(const uint8_t TotalPinsNumber, const uint8_t *pPinType)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  AllPinsData.OutputPinsNumber = AllPinsData.InputPinsNumber = 0;
  AllPinsData.CapNumber = AllPinsData.CapSeqCounter = 0;
  CounterChannelNumber = 0;
  SWQEncoderChannelNumber = 0;
  PeriodChannelNumber = 0;
  LedcChannelNumber = 0;
  ServoChannelNumber = 0;
  PwmChannelNumber = 0;
  pcnt_config_t dCounter;
  uint8_t UnusedPinNumber = 0;

  //  phase 1 : for every pin already programmed remove the actual function and restore the hard reset pin status 

  //  only if module was already programmed
  if (GetProgStatus() == 'Y')
  {
    //  check ahead if ADC24 master pin is going to be removed looping through all pPinType values looking for ADC_24
    bool ADC24MasterPinConfigured = false;
    for (uint8_t i = 0; i < TotalPinsNumber; i++)
    {
      if (pPinType[i] == ADC_24)
      {
        ADC24MasterPinConfigured = true;
      }
    }
    //  if adc24 master pin was configured and now is not configured
    if (IotModuleAdc.GetMasterPinConfigured() && !ADC24MasterPinConfigured)
    {
      // shut adc24 down : free SPI pins and reconfigure them as per hard reset
      IotModuleAdc.ShutDown();
    }

    //  for all pins
    for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
    {
      //  restore all used GPIO as input
      switch (AllPinsData.PinsParams[Index].Type)
      {
        case DIG_OUT:
        case PWM:
        case SERVO:
        case DIG_IN:
        case DIG_IN_PU:
        case ADC_16:
        case CAP_16:
        case COUNTER:
        case COUNTER_PU:
        case ENCODER_B:
        case ENCODER_B_PU:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
          ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, ModuleProps.PinsProps[Index].ResetMode));
          break;
        case DAC_8:
          //  disable the pin DAC functionality
          ESP_ERROR_CHECK(dac_output_disable((dac_channel_t)ModuleProps.PinsProps[Index].Channel1));
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
          ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, ModuleProps.PinsProps[Index].ResetMode));
          break;
        case ENCODER_A:
        case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 1
          //  detach SWQEncoder
          IotModuleOth.SWQEncoderDetach(AllPinsData.PinsParams[Index].Channel2);
#endif
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
          ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, ModuleProps.PinsProps[Index].ResetMode));
          break;
        case PERIOD:
        case PERIOD_PU:
          //  detach Period
          IotModuleOth.PeriodDetach(AllPinsData.PinsParams[Index].Channel2);
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
          ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, ModuleProps.PinsProps[Index].ResetMode));
          break;
        default:
          //  do nothing for other types
          break;
      }
      AllPinsData.PinsParams[Index].fValue = 0.0f;
    }
  }

  //  phase 2 : for every pin set the pin type according to the requested function

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - TotalPinsNumber : %d", __FILENAME__, __LINE__, TotalPinsNumber);
  for (uint8_t i = 0; i < TotalPinsNumber; i++)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Type : %d", __FILENAME__, __LINE__, pPinType[i]);
  }

  //  if new total within limits
  if (TotalPinsNumber <= AllocatedPinsNumber)
  {
    //  for all pins
    for (uint8_t Index = 0; Index < TotalPinsNumber; Index++)
    {
      //  set the programmed name
      AllPinsData.PinsParams[Index].Name = ModuleProps.PinsProps[Index].Name;
      //  get type
      PIN_TYPE Type = (PIN_TYPE)pPinType[Index];
      //  Check if this type is one of the possible types the pin can have
      bool ValidType = false;
      for (uint8_t i = 0; i < ModuleProps.PinsProps[Index].TypesSize; i++)
      {
        if (Type == ModuleProps.PinsProps[Index].Types[i])
        {
          ValidType = true;
          break;
        }
      }
      //  if not valid we set it unused
      if (!ValidType)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not valid - pin name : %d - type : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].Name, Type);
        Type = UNUSED;
      }
      //  set the type
      AllPinsData.PinsParams[Index].Type = Type;
      switch (Type)
      {
        case UNUSED:
          UnusedPinNumber++;
          break;

        case DIG_OUT:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_OUTPUT));
          ESP_ERROR_CHECK(gpio_set_level((gpio_num_t)AllPinsData.PinsParams[Index].Name, AllPinsData.PinsParams[Index].fValue == 0.0f ? 0 : 1));
          AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
          AllPinsData.OutputPinsNumber++;
          break;

        case PWM:
        case SERVO:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          if (LedcChannelNumber < MAX_LEDC_CHANNELS)
          {
            // -------------------------------------------------------------------
            //  PWM and SERVO are all configured in the InitPwmAndServo function
            // -------------------------------------------------------------------
            AllPinsData.PinsParams[Index].Channel2 = LedcChannelNumber;
            AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
            AllPinsData.OutputPinsNumber++;
            LedcChannelNumber++;
            if (Type == SERVO) ServoChannelNumber++;
            else PwmChannelNumber++;
          }
          else
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ledc error - LedcChannelNumber overflow", __FILENAME__, __LINE__);
          }
          break;

        case DAC_8:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(dac_output_enable((dac_channel_t)ModuleProps.PinsProps[Index].Channel1));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - DAC_8 - channel : %d", __FILENAME__, __LINE__, ModuleProps.PinsProps[Index].Channel1);
          AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
          AllPinsData.OutputPinsNumber++;
          break;

        case DIG_IN:
        case DIG_IN_PU:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
          switch (Type)
          {
            case DIG_IN:
              ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_FLOATING));
              break;
            case DIG_IN_PU:
              ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_PULLUP_ONLY));
              break;
            default:
              break;
          }
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_16:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          AllPinsData.PinsParams[Index].Channel2 = ModuleProps.PinsProps[Index].Channel1;
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case CAP_16:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          AllPinsData.PinsParams[Index].CapSeq = AllPinsData.CapNumber;
          AllPinsData.CapNumber++;
          break;

        case COUNTER:
        case COUNTER_PU:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          if (CounterChannelNumber < MAX_COUNTER_CHANNELS)
          {
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
            switch (Type)
            {
              case COUNTER:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_FLOATING));
                break;
              case COUNTER_PU:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_PULLUP_ONLY));
                break;
              default:
                break;
            }
            dCounter.pulse_gpio_num = AllPinsData.PinsParams[Index].Name;
            dCounter.ctrl_gpio_num = PCNT_PIN_NOT_USED;
            dCounter.unit = (pcnt_unit_t)CounterChannelNumber;  //  unit
            dCounter.channel = PCNT_CHANNEL_0;                  //  channel
            dCounter.pos_mode = PCNT_COUNT_INC;                 //  count Only On Rising-Edges
            dCounter.neg_mode = PCNT_COUNT_DIS;                 //  discard Falling-Edge
            dCounter.lctrl_mode = PCNT_MODE_KEEP;
            dCounter.hctrl_mode = PCNT_MODE_KEEP;
            dCounter.counter_h_lim = 0;
            dCounter.counter_l_lim = 0;

            ESP_ERROR_CHECK(pcnt_unit_config(&dCounter));
            //  pcnt_set_filter_value((pcnt_unit_t)CounterChannelNumber, 250);
            //  pcnt_filter_enable((pcnt_unit_t)CounterChannelNumber);
            pcnt_counter_pause((pcnt_unit_t)CounterChannelNumber);  //  pause
            pcnt_counter_clear((pcnt_unit_t)CounterChannelNumber);  //  clear
            pcnt_counter_resume((pcnt_unit_t)CounterChannelNumber); //  resume
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
            AllPinsData.PinsParams[Index].Channel2 = CounterChannelNumber;
            AllPinsData.InputPinsNumber++;
            CounterChannelNumber++;
          }
          else
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Counters error - CounterChannelNumber overflow", __FILENAME__, __LINE__);
          }
          break;

        case PERIOD:
        case PERIOD_PU:
          if (PeriodChannelNumber < cIotModuleOth::MAX_PERIOD_CHANNELS)
          {
            ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
            switch (Type)
            {
              case PERIOD:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_FLOATING));
                break;
              case PERIOD_PU:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_PULLUP_ONLY));
                break;
              default:
                break;
            }
            IotModuleOth.PeriodAttach(PeriodChannelNumber, ModuleProps.PinsProps[Index].Name);
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
            AllPinsData.PinsParams[Index].Channel2 = PeriodChannelNumber;
            AllPinsData.InputPinsNumber++;
            PeriodChannelNumber++;
          }
          else
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Periods error - PeriodChannelNumber overflow", __FILENAME__, __LINE__);
          }
          break;

        case ENCODER_A:
        case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
          if (CounterChannelNumber < MAX_COUNTER_CHANNELS)
#elif IOTMODULE_ENCODERS_TYPE == 1
          if (SWQEncoderChannelNumber < cIotModuleOth::MAX_SWQENCODER_CHANNELS)
#endif
          {
            //  at the moment the A and B GPIO pin setup is done at once when A is setup
            //  therefore A must be set before B and A and B must be consecutive
            //  AllPinsPropsFull's Name used because AllProgValues's Name is not programmed yet for chan B
            dCounter.pulse_gpio_num = ModuleProps.PinsProps[Index].Name;     //  rotary encoder chan A
            dCounter.ctrl_gpio_num = ModuleProps.PinsProps[Index + 1].Name;  //  rotary encoder chan B
            //  set the programmed names
            AllPinsData.PinsParams[Index].Name = dCounter.pulse_gpio_num;
            AllPinsData.PinsParams[Index + 1].Name = dCounter.ctrl_gpio_num;
            ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)dCounter.pulse_gpio_num));
            ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)dCounter.ctrl_gpio_num));
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)dCounter.pulse_gpio_num, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)dCounter.ctrl_gpio_num, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
            switch (Type)
            {
              case ENCODER_A:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)dCounter.pulse_gpio_num, GPIO_FLOATING));
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)dCounter.ctrl_gpio_num, GPIO_FLOATING));
                //  set the next pin type
                AllPinsData.PinsParams[Index + 1].Type = ENCODER_B;
                break;
              case ENCODER_A_PU:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)dCounter.pulse_gpio_num, GPIO_PULLUP_ONLY));
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)dCounter.ctrl_gpio_num, GPIO_PULLUP_ONLY));
                //  set the next pin type
                AllPinsData.PinsParams[Index + 1].Type = ENCODER_B_PU;
                break;
              default:
                break;
            }
#if IOTMODULE_ENCODERS_TYPE == 0
            dCounter.unit = (pcnt_unit_t)CounterChannelNumber;  //  unit
            dCounter.channel = PCNT_CHANNEL_0;                  //  channel
            dCounter.pos_mode = PCNT_COUNT_DEC;                 //  Decrement on Rising-Edges counts x 2
            dCounter.neg_mode = PCNT_COUNT_INC;                 //  Increment on Falling-Edge counts x 2
            dCounter.lctrl_mode = PCNT_MODE_KEEP;               //  rising A on low B = CW Step
            dCounter.hctrl_mode = PCNT_MODE_REVERSE;            //  rising A on high B = CCW Step
            dCounter.counter_h_lim = 0;
            dCounter.counter_l_lim = 0;

            ESP_ERROR_CHECK(pcnt_unit_config(&dCounter));
            //  pcnt_set_filter_value((pcnt_unit_t)CounterChannelNumber, 250);
            //  pcnt_filter_enable((pcnt_unit_t)CounterChannelNumber);
            pcnt_counter_pause((pcnt_unit_t)CounterChannelNumber);  //  pause
            pcnt_counter_clear((pcnt_unit_t)CounterChannelNumber);  //  clear
            pcnt_counter_resume((pcnt_unit_t)CounterChannelNumber); //  resume
#elif IOTMODULE_ENCODERS_TYPE == 1
            IotModuleOth.SWQEncoderAttach(SWQEncoderChannelNumber, ModuleProps.PinsProps[Index].Name, ModuleProps.PinsProps[Index + 1].Name);
#endif
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
#if IOTMODULE_ENCODERS_TYPE == 0
            AllPinsData.PinsParams[Index].Channel2 = CounterChannelNumber;
#elif IOTMODULE_ENCODERS_TYPE == 1
            AllPinsData.PinsParams[Index].Channel2 = SWQEncoderChannelNumber;
#endif
            AllPinsData.InputPinsNumber++;
#if IOTMODULE_ENCODERS_TYPE == 0
            CounterChannelNumber++;
#elif IOTMODULE_ENCODERS_TYPE == 1
            SWQEncoderChannelNumber++;
#endif
          }
          else
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Encoders error - CounterChannelNumber overflow", __FILENAME__, __LINE__);
          }
          break;

        case ENCODER_B:
        case ENCODER_B_PU:
          //  encoder_b dummy pin
          UnusedPinNumber++;
          break;

        case GEN_OUT_8:
        case GEN_OUT_16:
        case GEN_OUT_24:
        case GEN_OUT_FLOAT:
          AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
          AllPinsData.OutputPinsNumber++;
          break;

        case GEN_IN_8:
        case GEN_IN_16:
        case GEN_IN_24:
        case GEN_IN_FLOAT:
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_24:
          //  if adc24 configured and master pin not yet configured
          if (!IotModuleAdc.GetMasterPinConfigured())
          {
            //  configure master pin
            IotModuleAdc.ConfigureMasterPin();
          }
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_24_CH:
          //  adc24 pin used for conversion value
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_24_CH_B:
          //  adc24 dummy pin
          UnusedPinNumber++;
          break;

        default:
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not handled - Type : %d", __FILENAME__, __LINE__, Type);
          break;
      }
    }

    AllPinsData.TotalPinsNumber = TotalPinsNumber;

    if ((AllPinsData.OutputPinsNumber + AllPinsData.InputPinsNumber + UnusedPinNumber) == TotalPinsNumber)
    {
      //  PWM and SERVO configuration
      InitPwmAndServo();
      //	all module pins programmed
      ProgStatus =  static_cast<PROG_STATUS>(ProgStatus | MODULE);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - all module pins programmed : %d", __FILENAME__, __LINE__, TotalPinsNumber);
      for (uint8_t i = 0; i < TotalPinsNumber; i++)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Name : %d, Type : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[i].Name, AllPinsData.PinsParams[i].Type);
      }
    }
    else
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin mismatch", __FILENAME__, __LINE__);
      //  some pin programmed and some not, pin program Mismatch
      ProgStatus = NOTHING;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin allocation error", __FILENAME__, __LINE__);
    //  no pin programmed
    AllPinsData.TotalPinsNumber = 0;
    ProgStatus = NOTHING;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  PWM and SERVO functions
// ----------------------------------------------------------------------------------------------------
//  PinName  = ESP32 pin mames                                 ( 36 / 39 / 25 / 26  ...  19 / 22 / 21 )
//  PinIndex = pin numeration as ordered in the IotHAL         (  0 /  1 /  2 /  3  ...  26 / 27 / 28 )
//  pwm_ch   = 16 pwm channels ordered as the hal progression  ( 0 to 15 )
//  ledc_ch  = 16 pwm channels coupled with the same frequency ( 0-1 / 2-3 / 4-5  ...  12-13 / 14-15 )
// ----------------------------------------------------------------------------------------------------
//  led_ch  0 and  1 have the same frequency and are served by the pwm timer 1
//  led_ch  2 and  3 have the same frequency and are served by the pwm timer 2
//  led_ch  4 and  5 have the same frequency and are served by the pwm timer 3
//  led_ch  6 and  7 have the same frequency and are served by the pwm timer 4
//  led_ch  8 and  9 have the same frequency and are served by the pwm timer 5  (slow speed)
//  led_ch 10 and 11 have the same frequency and are served by the pwm timer 6  (slow speed)
//  led_ch 12 and 13 have the same frequency and are served by the pwm timer 7  (slow speed)
//  led_ch 14 and 15 have the same frequency and are served by the pwm timer 8  (slow speed)
// ----------------------------------------------------------------------------------------------------
//  High-speed channels parameters are controlled by the hardware itself
//  Slow-speed channels are corrected by the library in ledc_timer_config() and ledc_timer_set()
// ----------------------------------------------------------------------------------------------------
//  The firmware for the slow timers has a bug.
//  Setting Width to zero, the timers are not updated and continue to use the old Width.
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::InitPwmAndServo(void)
{
  //  init with invalid values
  for (uint8_t i = 0; i < MAX_LEDC_CHANNELS; i++) AllPwmServoData.Params[i].ledc_ch = 255;
  for (uint8_t i = 0; i < AllPinsData.TotalPinsNumber; i++) AllPinsData.PinsParams[i].PwmChannel = 255;
  // -----------------------------------------------------------
  uint8_t pwm_ch = 0;
  uint8_t ledc_ch = 0;
  //  first pass for SERVO
  for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
  {
    switch (AllPinsData.PinsParams[Index].Type)
    {
      case SERVO:
        //  fixed frequency 50 Hz
        PwmInitChannel(AllPinsData.PinsParams[Index].Name, pwm_ch, ledc_ch);
        AllPinsData.PinsParams[Index].PwmChannel = pwm_ch;
        pwm_ch++;
        ledc_ch++;
        break;
      default:
        break;
    }
  }

  // --------------------------------------------------------------------------------
  //  If the SERVO are not even, avoid to couple the first PWM with the last SERVO
  //  If ledc_ch = 1, 3, 5, 7, 9, 11, 13 increase it to 2, 4, 6, 8, 10, 12, 14
  // --------------------------------------------------------------------------------
  if (ledc_ch % 2 == 1) ledc_ch++;

  //  second pass for PWM
  for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
  {
    switch (AllPinsData.PinsParams[Index].Type)
    {
      case PWM:
        //  variable frequency from 0.02 Hz to 4 MHz
        PwmInitChannel(AllPinsData.PinsParams[Index].Name, pwm_ch, ledc_ch);
        AllPinsData.PinsParams[Index].PwmChannel = pwm_ch;
        pwm_ch++;
        //  PWM are coupled if : ServoCount + PwmCount * 2 is more than 16
        if (ServoChannelNumber + PwmChannelNumber * 2 > 16) ledc_ch++;
        else ledc_ch += 2;
        break;
      default:
        break;
    }
  }
}

void cIotModuleApp::PwmInitChannel(const uint8_t PinName, const uint8_t pwm_ch, const uint8_t ledc_ch)
{
  if (pwm_ch > MAX_LEDC_CHANNELS - 1) return;
  if (ledc_ch > MAX_LEDC_CHANNELS - 1) return;
  //  fill the pwmChannel array with the ledc_ch
  AllPwmServoData.Params[pwm_ch].ledc_ch = ledc_ch;
  //  attach the PinName to ledc_ch
  ledcAttachPin(PinName, ledc_ch);
  //  set a default frequency to init a valid MaxWidth
  PwmSetFrequency(pwm_ch, 50);
  //  debug
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - pinName : %d - pwm_ch : %d - ledc_ch : %d", __FILENAME__, __LINE__, PinName, pwm_ch, ledc_ch);
}

void cIotModuleApp::RPCPwmSetFrequency(const uint8_t PinIndex, const float Frequency)
{
  uint8_t pwm_ch = AllPinsData.PinsParams[PinIndex].PwmChannel;
  PwmSetFrequency(pwm_ch, Frequency);
  //  debug
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - PinIndex : %d - pwm_ch : %d - freq : %f", __FILENAME__, __LINE__, PinIndex, pwm_ch, Frequency);
}

void cIotModuleApp::PwmSetFrequency(const uint8_t pwm_ch, float Frequency)
{
  if (pwm_ch > MAX_LEDC_CHANNELS - 1) return;
  //  ledc-channels 1,3,5,7,9,11,13,15 have the same freq of 0,2,4,6,8,10,12,14
  if (Frequency < 0.02) Frequency = 0.02;
  if (Frequency > 40000000) Frequency = 40000000;
  //  calculate the resolution
  uint32_t MaxWidth = 1048576; // <-- 2^20
  uint8_t ResBits = 20;
  while (ResBits >= 2)
  {
    if ((uint32_t)(MaxWidth * Frequency) < 80000000) break;
    ResBits -= 1;
    MaxWidth /= 2;
  }
  //  setup the MaxWidth
  AllPwmServoData.Params[pwm_ch].MaxWidth = MaxWidth;
  //  setup the ledc-ch frequency and resolution
  uint8_t ledc_ch = AllPwmServoData.Params[pwm_ch].ledc_ch;
  ledcSetup(ledc_ch, Frequency, ResBits);
  //  setup also the associated ledc-ch 8 to 15
  ledcSetup(ledc_ch + 1, Frequency, ResBits);
  //  init the PwmWidth to 0%
  PwmSetWidth(pwm_ch, 0);
  //  debug
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - pwm_ch : %d - ledc_ch : %d - Frequency : %f", __FILENAME__, __LINE__, pwm_ch, ledc_ch, Frequency);
}

void cIotModuleApp::PwmSetWidth(const uint8_t pwm_ch, const float Width)
{
  uint8_t ledc_ch = AllPwmServoData.Params[pwm_ch].ledc_ch;
  if (ledc_ch > MAX_LEDC_CHANNELS - 1) return;

  uint32_t PwmWidth = Width * AllPwmServoData.Params[pwm_ch].MaxWidth + 0.5;
  // -------------------------------------------------------------
  // workaround against library bug
  // (slow-speed pwm timers not updated when PwmWidth is zero)
  if (ledc_ch >= 8) if (PwmWidth < 1) PwmWidth = 1;
  // -------------------------------------------------------------
  ledcWrite(ledc_ch, PwmWidth);
  //  debug
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - pwm_ch : %d - ledc_ch : %d - PwmWidth : %d", __FILENAME__, __LINE__, pwm_ch, ledc_ch, PwmWidth);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function copy the output values from the communication buffer
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::RPCSetOutputXchgValues(const uint8_t IndexCount, const float *pOutValue)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  if (IndexCount < AllPinsData.OutputPinsNumber)
  {
    uint8_t Index = AllPinsData.OutputIndexes[IndexCount];
    switch (AllPinsData.PinsParams[Index].Type)
    {
      case SERVO:
      case PWM:
      case DIG_OUT:
      case DAC_8:
      case GEN_OUT_FLOAT:
      case GEN_OUT_8:
      case GEN_OUT_16:
      case GEN_OUT_24:
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - %d", __FILENAME__, __LINE__, Index);
        AllPinsData.PinsParams[Index].fValue = *pOutValue;
        break;
      default:
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not handled", __FILENAME__, __LINE__);
        //  zero value
        AllPinsData.PinsParams[Index].fValue = 0.0f;
        //  pin program mismatch
        ProgStatus = NOTHING;
        break;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin mismatch", __FILENAME__, __LINE__);
    ProgStatus = NOTHING;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function copy the input values to the communication buffer
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::RPCGetInputXchgValues(const uint8_t IndexCount, float *pInValue)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  if (IndexCount < AllPinsData.InputPinsNumber)
  {
    uint8_t Index = AllPinsData.InputIndexes[IndexCount];
    switch (AllPinsData.PinsParams[Index].Type)
    {
      case DIG_IN:
      case DIG_IN_PU:
      case ADC_16:
      case CAP_16:
      case COUNTER:
      case COUNTER_PU:
      case PERIOD:
      case PERIOD_PU:
      case ENCODER_A:
      case ENCODER_A_PU:
      case ENCODER_B:
      case ENCODER_B_PU:
      case GEN_IN_8:
      case GEN_IN_16:
      case GEN_IN_24:
      case GEN_IN_FLOAT:
      case ADC_24:
      case ADC_24_CH:
        *pInValue = AllPinsData.PinsParams[Index].fValue;
        break;
      default:
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not handled - Type : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].Type);
        //  zero value
        *pInValue = 0.0f;
        //  pin program Mismatch
        ProgStatus = NOTHING;
        break;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin number mismatch", __FILENAME__, __LINE__);
    ProgStatus = NOTHING;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                 User program interface functions
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::GenericWrite8 (const uint8_t PinName, const uint8_t Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

void cIotModuleApp::GenericWrite16(const uint8_t PinName, const uint16_t Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

void cIotModuleApp::GenericWrite24(const uint8_t PinName, const uint32_t Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

void cIotModuleApp::GenericWriteFloat(const uint8_t PinName, const float Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

uint8_t cIotModuleApp::GenericRead8 (const uint8_t PinName)
{
  uint8_t RetValue = 0;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}

uint16_t cIotModuleApp::GenericRead16(const uint8_t PinName)
{
  uint16_t RetValue = 0;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}

uint32_t cIotModuleApp::GenericRead24(const uint8_t PinName)
{
  uint32_t RetValue = 0;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}

float cIotModuleApp::GenericReadFloat(const uint8_t PinName)
{
  float RetValue = 0.0f;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}
