//================================================================================================
//                                  cIotModuleOth body
//================================================================================================

cIotModuleOth::cIotModuleOth()
{
}

void cIotModuleOth::Initialize(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_SWQENCODER_CHANNELS; ChannelNumber++)
  {
    SWQEncodersData.Params[ChannelNumber].Attached = false;
  }
  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_PERIOD_CHANNELS; ChannelNumber++)
  {
    PeriodsData.Params[ChannelNumber].Attached = false;
  }
}

void cIotModuleOth::PreFuota(void)
{
  //  Stop any interrupt activity
  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_SWQENCODER_CHANNELS; ChannelNumber++)
  {
    if (SWQEncodersData.Params[ChannelNumber].Attached)
    {
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)SWQEncodersData.Params[ChannelNumber].PinA_Name));
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)SWQEncodersData.Params[ChannelNumber].PinB_Name));
      SWQEncodersData.Params[ChannelNumber].Attached = false;
    }
  }
  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_PERIOD_CHANNELS; ChannelNumber++)
  {
    if (PeriodsData.Params[ChannelNumber].Attached)
    {
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)PeriodsData.Params[ChannelNumber].PinName));
      PeriodsData.Params[ChannelNumber].Attached = false;
    }
  }
}

bool cIotModuleOth::SWQEncoderAttach(const uint8_t Channel, const uint8_t PinA_Name, const uint8_t PinB_Name)
{
  bool ExitValue = false;
  if (Channel < MAX_SWQENCODER_CHANNELS)
  {
    if (SWQEncodersData.Params[Channel].Attached == false)
    {
      //  store params
      SWQEncodersData.Params[Channel].PinA_Name = PinA_Name;
      SWQEncodersData.Params[Channel].PinB_Name = PinB_Name;
      //  init phase
      SWQEncodersData.Params[Channel].PinAB_StatusOld = gpio_get_level((gpio_num_t)PinB_Name);
      SWQEncodersData.Params[Channel].PinAB_StatusOld <<= 1;
      SWQEncodersData.Params[Channel].PinAB_StatusOld += gpio_get_level((gpio_num_t)PinA_Name);
      SWQEncodersData.Params[Channel].PinAB_StatusOld &= 0b11;
      //  reset count
      SWQEncodersData.Params[Channel].QuadratureCount = 0;
      //  ok
      SWQEncodersData.Params[Channel].Attached = true;
      //  enable GPIO interrupt on change
      ESP_ERROR_CHECK(gpio_set_intr_type((gpio_num_t)PinA_Name, GPIO_INTR_ANYEDGE));
      ESP_ERROR_CHECK(gpio_set_intr_type((gpio_num_t)PinB_Name, GPIO_INTR_ANYEDGE));
      //  add ISR handler
      void (* ISR_SWQEncoder_OnChange)(void *arg);
      switch (Channel)
      {
        case 0:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_0;
          break;
        case 1:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_1;
          break;
        case 2:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_2;
          break;
        case 3:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_3;
          break;
        case 4:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_4;
          break;
        case 5:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_5;
          break;
        case 6:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_6;
          break;
        case 7:
          ISR_SWQEncoder_OnChange = ISR_SWQEncoderOnChange_7;
          break;
      }
      //  add ISR Handler
      ESP_ERROR_CHECK(gpio_isr_handler_add((gpio_num_t)PinA_Name, ISR_SWQEncoder_OnChange, this));
      ESP_ERROR_CHECK(gpio_isr_handler_add((gpio_num_t)PinB_Name, ISR_SWQEncoder_OnChange, this));
      //  enable GPIO interrupt on change
      //  ESP_ERROR_CHECK(gpio_intr_enable((gpio_num_t)PinA_Name));
      //  ESP_ERROR_CHECK(gpio_intr_enable((gpio_num_t)PinB_Name));
      ExitValue = true;
    }
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d - PinA_Name : %d - PinB_Name : %d", __FILENAME__, __LINE__, ExitValue, Channel, PinA_Name, PinB_Name);
  }
  return ExitValue;
}

bool cIotModuleOth::SWQEncoderDetach(const uint8_t Channel)
{
  bool ExitValue = false;
  if (Channel < MAX_SWQENCODER_CHANNELS)
  {
    if (SWQEncodersData.Params[Channel].Attached == true)
    {
      //  disable GPIO interrupt on change
      //  ESP_ERROR_CHECK(gpio_intr_disable((gpio_num_t)SWQEncodersData.Params[Channel].PinA_Name));
      //  ESP_ERROR_CHECK(gpio_intr_disable((gpio_num_t)SWQEncodersData.Params[Channel].PinB_Name));
      //  remove ISR Handler
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)SWQEncodersData.Params[Channel].PinA_Name));
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)SWQEncodersData.Params[Channel].PinB_Name));

      SWQEncodersData.Params[Channel].Attached = false;
      ExitValue = true;
    }
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}

bool cIotModuleOth::SWQEncoderGetQuadratureCount(const uint8_t Channel, int16_t *QuadratureCount)
{
  bool ExitValue = false;
  if (Channel < MAX_SWQENCODER_CHANNELS)
  {
    if (SWQEncodersData.Params[Channel].Attached)
    {
      *QuadratureCount = SWQEncodersData.Params[Channel].QuadratureCount;
      ExitValue = true;
    }
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}

void IRAM_ATTR cIotModuleOth::SWQEncoderOnChangeCommon(void *arg, const uint8_t Channel)
{
  cIotModuleOth *pIotModuleOth = (cIotModuleOth *)arg;
  //  bit 3 = B(old), bit 2 = A(old), bit 1 = B(new), bit 0 = A(new)
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_StatusOld <<= 2;
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status = gpio_get_level((gpio_num_t)pIotModuleOth->SWQEncodersData.Params[Channel].PinB_Name);
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status <<= 1;
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status += gpio_get_level((gpio_num_t)pIotModuleOth->SWQEncodersData.Params[Channel].PinA_Name);
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status |= pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_StatusOld;
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status &= 0b1111;
  //  read increment/decrememt/noaction from lookup table
  pIotModuleOth->SWQEncodersData.Params[Channel].QuadratureCount += pIotModuleOth->QuadratureTable[pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status];
  //  store new data
  pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_StatusOld = pIotModuleOth->SWQEncodersData.Params[Channel].PinAB_Status;
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_0(void *arg)
{
  //  channel 0
  SWQEncoderOnChangeCommon(arg, 0);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_1(void *arg)
{
  //  channel 1
  SWQEncoderOnChangeCommon(arg, 1);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_2(void *arg)
{
  //  channel 2
  SWQEncoderOnChangeCommon(arg, 2);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_3(void *arg)
{
  //  channel 3
  SWQEncoderOnChangeCommon(arg, 3);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_4(void *arg)
{
  //  channel 4
  SWQEncoderOnChangeCommon(arg, 4);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_5(void *arg)
{
  //  channel 5
  SWQEncoderOnChangeCommon(arg, 5);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_6(void *arg)
{
  //  channel 6
  SWQEncoderOnChangeCommon(arg, 6);
}

void IRAM_ATTR cIotModuleOth::ISR_SWQEncoderOnChange_7(void *arg)
{
  //  channel 7
  SWQEncoderOnChangeCommon(arg, 7);
}

//  period

bool cIotModuleOth::PeriodAttach(const uint8_t Channel, const uint8_t PinName)
{
  bool ExitValue = false;
  if (Channel < MAX_PERIOD_CHANNELS)
  {
    if (PeriodsData.Params[Channel].Attached == false)
    {
      //  store params
      PeriodsData.Params[Channel].PinName = PinName;
      //  reset count
      PeriodsData.Params[Channel].Count = 0;
      //  init timer register old
      PeriodsData.Params[Channel].TimeCounterOld = 0;
      //  ok
      PeriodsData.Params[Channel].Attached = true;
      //  enable GPIO interrupt on rise
      ESP_ERROR_CHECK(gpio_set_intr_type((gpio_num_t)PinName, GPIO_INTR_POSEDGE ));
      //  add ISR handler
      void (* ISR_Period_OnRise)(void *arg);
      switch (Channel)
      {
        case 0:
          ISR_Period_OnRise = ISR_PeriodOnRise_0;
          break;
        case 1:
          ISR_Period_OnRise = ISR_PeriodOnRise_1;
          break;
        case 2:
          ISR_Period_OnRise = ISR_PeriodOnRise_2;
          break;
        case 3:
          ISR_Period_OnRise = ISR_PeriodOnRise_3;
          break;
        case 4:
          ISR_Period_OnRise = ISR_PeriodOnRise_4;
          break;
        case 5:
          ISR_Period_OnRise = ISR_PeriodOnRise_5;
          break;
        case 6:
          ISR_Period_OnRise = ISR_PeriodOnRise_6;
          break;
        case 7:
          ISR_Period_OnRise = ISR_PeriodOnRise_7;
          break;
      }
      //  add ISR Handler
      ESP_ERROR_CHECK(gpio_isr_handler_add((gpio_num_t)PinName, ISR_Period_OnRise, this));
      //  enable GPIO interrupt on change
      //  ESP_ERROR_CHECK(gpio_intr_enable((gpio_num_t)PinName));
      ExitValue = true;
    }
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d - PinName : %d", __FILENAME__, __LINE__, ExitValue, Channel, PinName);
  }
  return ExitValue;
}

bool cIotModuleOth::PeriodDetach(const uint8_t Channel)
{
  bool ExitValue = false;
  if (Channel < MAX_PERIOD_CHANNELS)
  {
    if (PeriodsData.Params[Channel].Attached == true)
    {
      //  disable GPIO interrupt on change
      //  ESP_ERROR_CHECK(gpio_intr_disable((gpio_num_t)PeriodsData.Params[Channel].PinName));
      //  remove ISR Handler
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)PeriodsData.Params[Channel].PinName));

      PeriodsData.Params[Channel].Attached = false;
      ExitValue = true;
    }
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}

bool cIotModuleOth::PeriodGetCount(const uint8_t Channel, uint32_t *Count)
{
  bool ExitValue = false;
  if (Channel < MAX_PERIOD_CHANNELS)
  {
    if (PeriodsData.Params[Channel].Attached == true)
    {
      //  read the period count
      *Count = PeriodsData.Params[Channel].Count;
      ExitValue = true;
    }
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}

void IRAM_ATTR cIotModuleOth::PeriodOnRiseCommon(void *arg, const uint8_t Channel)
{
  cIotModuleOth *pIotModuleOth = (cIotModuleOth *)arg;
  //  value of timer register at interrupt
  //  *** IotHAL processing must be modified according to the frequency, now 240 MHz base frequency ***

  //  first option (1 MHz base frequency):
  //  uint32_t TimeCounter = micros() * 10;
  //  second option (XTHAL_GET_CCOUNT macro and inline code are exactly the same code), base frequency 240 Mhz:
  uint32_t TimeCounter = XTHAL_GET_CCOUNT();
  //  uint32_t TimeCounter = ({ int __ccount; __asm__ __volatile__("rsr.ccount %0" : "=a"(__ccount)); __ccount; });

  //  period count between rising edges
  pIotModuleOth->PeriodsData.Params[Channel].Count = TimeCounter - pIotModuleOth->PeriodsData.Params[Channel].TimeCounterOld;
  //  store latest reading as start for next calculation
  pIotModuleOth->PeriodsData.Params[Channel].TimeCounterOld = TimeCounter;
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_0(void *arg)
{
  //  channel 0
  PeriodOnRiseCommon(arg, 0);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_1(void *arg)
{
  //  channel 1
  PeriodOnRiseCommon(arg, 1);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_2(void *arg)
{
  //  channel 2
  PeriodOnRiseCommon(arg, 2);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_3(void *arg)
{
  //  channel 3
  PeriodOnRiseCommon(arg, 3);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_4(void *arg)
{
  //  channel 4
  PeriodOnRiseCommon(arg, 4);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_5(void *arg)
{
  //  channel 5
  PeriodOnRiseCommon(arg, 5);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_6(void *arg)
{
  //  channel 6
  PeriodOnRiseCommon(arg, 6);
}

void IRAM_ATTR cIotModuleOth::ISR_PeriodOnRise_7(void *arg)
{
  //  channel 7
  PeriodOnRiseCommon(arg, 7);
}
